#! /usr/bin/env python
# -*- coding: utf-8 -*-
####################
# Indigo Tuya plugin
# Copyright (c) 2023 Alan Carter. All rights reserved.
#

################################################################################
# Imports
################################################################################

import ipaddress
import json
import datetime
import base64
import colorsys
from tuya_iot import (
    TuyaOpenAPI,
    AuthType
)
import tinytuya


################################################################################
# Globals
################################################################################

def updateVar(name, value, folder=0):

    if name not in indigo.variables:
        indigo.variable.create(name, value=value, folder=folder)
    else:
        indigo.variable.updateValue(name, value)

################################################################################
class Plugin(indigo.PluginBase):

    ########################################
    # Class properties
    ########################################

    def __init__(self, pluginId, pluginDisplayName, pluginVersion, pluginPrefs):
     
        indigo.PluginBase.__init__(self, pluginId, pluginDisplayName, pluginVersion, pluginPrefs)
        self.userName = pluginPrefs.get("userName", "")
        self.userPass = pluginPrefs.get("userPass", "")
        self.clientId = pluginPrefs.get("clientId", "")
        self.clientSecret = pluginPrefs.get("clientSecret", "")
        self.endPoint = pluginPrefs.get("endPoint", "")
        self.showEnergyInNotes = pluginPrefs.get("energyNotes", False)
        self.logComms = pluginPrefs.get("logComms", False)
        self.logCommands = pluginPrefs.get("logCommands", False)
        self.powerRetries = 0
        self.power150Retries = 0
        self.evRetries = 0
        self.wcRetries = 0
        self.switchCommand = {}

    
    ########################################
    def __del__(self):
    
        indigo.PluginBase.__del__(self)
        
    ########################################
    def startup(self):
    
        indigo.server.log("Starting Tuya plugin")
        self.openapi = TuyaOpenAPI(self.endPoint, self.clientId, self.clientSecret, AuthType.SMART_HOME)
        self.openapi.get_token()

    def deviceStartComm(self, dev):
    
        dev.stateListOrDisplayStateIdChanged()
        return


    def shutdown(self):
    
        indigo.server.log("Stopping Tuya plugin")
        pass
  
    ########################################
    
    def runConcurrentThread(self):
    
        self.getCommands()
        indigo.server.log("Starting Tuya monitoring thread")
        try:
            while True:
                self.getReadings()
                self.sleep(20)
        except self.StopThread:
            self.shutdown()



    ########################################
    # Called when user prefs are changed
    ########################################
    
    def closedPrefsConfigUi(self, valuesDict, userCancelled):
        # If the user saved preferences, update logging parameters
        if userCancelled == False:

            energyNotes = self.showEnergyInNotes
            self.showEnergyInNotes = valuesDict.get("energyNotes", False)
            if self.showEnergyInNotes != energyNotes:
                if self.showEnergyInNotes:
                    indigo.server.log("Energy will be shown in Notes column")
                else:
                    indigo.server.log("Energy will not be shown in Notes column")
                    
            logComms = self.logComms
            self.logComms = valuesDict.get("logComms", False)
            if self.logComms != logComms:
                if self.logComms:
                    indigo.server.log("Comms logging is on")
                else:
                    indigo.server.log("Comms logging is off")

            logCommands = self.logCommands
            self.logCommands = valuesDict.get("logCommands", False)
            if self.logCommands != logCommands:
                if self.logCommands:
                    indigo.server.log("Command logging is on")
                else:
                    indigo.server.log("Command logging is off")
                    

    ###########################################
    # Called when a device is added or changed
    ###########################################
    
    def validateDeviceConfigUi(self, valuesDict, typeId, devId):
        # Force refresh of switch/bulb on/off commands when a new device is added
        self.getCommands()
        return True


        
    ########################################
    # Tuya specific functions
    ########################################
    
    def getCommands(self):
        # Find on/off command for plugs and bulbs
        
        for dev in indigo.devices.iter("self"):

            try:
                switchFunctions = self.openapi.get("/v1.0/iot-03/devices/{}/functions".format(dev.address))
                for i in range(0, len(switchFunctions["result"]["functions"])):
                    commandCode = switchFunctions["result"]["functions"][i]["code"]
                    if self.logComms:
                        indigo.server.log("----------------------------------------------------")
                        indigo.server.log("Found control function for %s:" % dev.name)
                        indigo.server.log(str(commandCode))
                    if ("switch" in commandCode) and (len(commandCode) < 12):
                        self.switchCommand[dev.address] = commandCode
            except:
                if self.logComms:
                    indigo.server.log("----------------------------------------------------")
                    indigo.server.log("No control functions available for %s" % dev.name)
                    indigo.server.log(str(switchFunctions))
                continue
    


    def getReadings(self):
    
        for dev in indigo.devices.iter("self"):

            # Get readings
            try:     
                response = self.openapi.get("/v1.0/iot-03/devices/{}/status".format(dev.address))
            except:
                indigo.server.log("Error accessing Tuya servers")
                continue

            if response["success"] == True:
                if self.logComms:
                    indigo.server.log("----------------------------------------------------")
                    indigo.server.log("Device status for %s" % dev.name)
                    indigo.server.log(str(response))
                    
                if dev.deviceTypeId == "tuyaTemp":
                    temperature = 0
                    humidity = 0
                    battery = 0
                    batteryState = ""
                    unit = ""
                    for i in range(0, len(response["result"])):
                        vName = response["result"][i]["code"]
                        vValue = response["result"][i]["value"]
                        if "temperature" in vName:
                            temperature = float(vValue) / 10
                        elif "humidity" in vName:
                            humidity = vValue
                        elif "battery_percentage" in vName:
                            battery = vValue
                        elif "battery_state" in vName:
                            batteryState = vValue
                        elif "temp_unit_convert" in vName:
                            unit = vValue
                            if unit == "f":
                                temperature = (temperature * 9 / 5) + 32
                    key_value_list = [
                        {'key':'temperature', 'value':temperature, 'uiValue':str(temperature)+" "+unit.upper(), 'decimalPlaces':1},
                        {'key':'humidity', 'value':float(humidity), 'uiValue':str(humidity)+"%"},
                        {'key':'battery', 'value':battery},
                        {'key':'battery_state', 'value':batteryState},
                    ]
                     
                elif dev.deviceTypeId == "tuyaPlug":
                    onState = False
                    current = 0
                    voltage = 0
                    power = 0
                    for i in range(0, len(response["result"])):
                        vName = response["result"][i]["code"]
                        vValue = response["result"][i]["value"]
                        if "switch" in vName:
                            onState = vValue
                        elif "current" in vName:
                            current = float(vValue) / 1000
                        elif "voltage" in vName:
                            voltage = float(vValue) / 10
                        elif "power" in vName:
                            power = float(vValue) / 10
                    if onState:
                        dev.updateStateOnServer(key="onOffState", value=True, clearErrorState=True)
                        dev.updateStateImageOnServer(indigo.kStateImageSel.PowerOn)
                    else:
                        dev.updateStateOnServer(key="onOffState", value=False, clearErrorState=True)
                        dev.updateStateImageOnServer(indigo.kStateImageSel.PowerOff)
                    if len(response['result']) < 2:
                        # Plug not reporting energy data
                        continue
                    key_value_list = [
                        {'key':'cur_current', 'value':current, 'uiValue':str(current)+" "+" A", 'decimalPlaces':1},
                        {'key':'cur_power', 'value':float(power), 'uiValue':str(power)+" W"},
                        {'key':'cur_voltage', 'value':float(voltage), 'uiValue':str(voltage)+" V"}
                    ]
                    if self.showEnergyInNotes:
                        dev.description = str(power)+" W"
                        dev.replaceOnServer()
                    elif dev.description != "":
                        dev.description = ""
                        dev.replaceOnServer()
                        
                elif dev.deviceTypeId == "tuyaBulb":
                    onState = False
                    mode = ""
                    brightness = 0
                    colourTemp = 0
                    for i in range(0, len(response["result"])):
                        vName = response["result"][i]["code"]
                        vValue = response["result"][i]["value"]
                        if "switch" in vName:
                            onState = vValue
                        elif "bright" in vName:
                            try:
                                brightness = float(vValue) / 10
                            except:
                                brightness = 0
                        elif "temp" in vName:
                            colourTemp = int(vValue)
                        elif "mode" in vName:
                            mode = vValue
                    if onState:
                        dev.updateStateOnServer(key="onOffState", value=True, clearErrorState=True)
                        dev.updateStateImageOnServer(indigo.kStateImageSel.PowerOn)
                    else:
                        dev.updateStateOnServer(key="onOffState", value=False, clearErrorState=True)
                        dev.updateStateImageOnServer(indigo.kStateImageSel.PowerOff)
                        brightness = 0
                    brightnessInt = int(brightness)
                    key_value_list = [
                        {'key':'mode', 'value':mode},
                        {'key':"brightnessLevel", 'value':brightnessInt, 'uiValue':str(brightnessInt)+"%"},
                        {'key':'colour_temp', 'value':colourTemp, 'uiValue':str(colourTemp)}
                    ]
                    
                elif dev.deviceTypeId == "tuyaPresence":
                    presence = ""
                    sensitivity = ""
                    near_detection = ""
                    far_detection = ""
                    target_dis_closest = ""
                    for i in range(0, len(response["result"])):
                        vName = response["result"][i]["code"]
                        vValue = response["result"][i]["value"]
                        if "presence" in vName:
                            presence = vValue
                        elif "sensitivity" in vName:
                            sensitivity = vValue
                        elif "near_detection" in vName:
                            near_detection = vValue
                        elif "far_detection" in vName:
                            far_detection = vValue
                        elif "target_dis_closest" in vName:
                            target_dis_closest = vValue
                    key_value_list = [
                        {'key':'presence_state', 'value':presence, 'uiValue':str(presence)},
                        {'key':'sensitivity', 'value':sensitivity, 'uiValue':str(sensitivity)},
                        {'key':'near_detection', 'value':near_detection, 'uiValue':str(near_detection)},
                        {'key':'far_detection', 'value':far_detection, 'uiValue':str(far_detection)},
                        {'key':'target_dis_closest', 'value':target_dis_closest, 'uiValue':str(target_dis_closest)}
                    ]
                    if presence == "presence":
                        dev.updateStateImageOnServer(indigo.kStateImageSel.MotionSensorTripped)
                    else:
                        dev.updateStateImageOnServer(indigo.kStateImageSel.MotionSensor)
                        
                elif dev.deviceTypeId == "tuyaVac":
                    key_value_list = []
                    for i in range (0, len(response["result"])):
                        key_value_list.append({'key':str(response["result"][i]["code"]), 'value':str(response["result"][i]["value"])})
                        
                elif dev.deviceTypeId == "tuyaSmoke":
                    key_value_list = []
                    for i in range (0, len(response["result"])):
                        if i == 3:
                            if "SupportsBatteryLevel" not in dev.pluginProps:
                                localPropsCopy = dev.pluginProps
                                localPropsCopy.update({"SupportsBatteryLevel":True})
                                dev.replacePluginPropsOnServer(localPropsCopy)
                            dev.updateStateOnServer(key = 'batteryLevel', value = int(response["result"][i]["value"]))
                            continue
                        if i == 4:
                            muted = response["result"][i]["value"]
                            if muted:
                                dev.setErrorStateOnServer("MUTED")
                            else:
                                dev.setErrorStateOnServer(None)
                        key_value_list.append({'key':str(response["result"][i]["code"]), 'value':str(response["result"][i]["value"])})
                        
                elif dev.deviceTypeId == "tuyaPower":
                    try:     
                        response = self.openapi.get("/v2.0/cloud/thing/{}/shadow/properties".format(dev.address))
                    except:
                        indigo.server.log("Error accessing Tuya servers for Smart Power Monitor")
                        continue
                    if self.logComms:
                        indigo.server.log("----------------------------------------------------")
                        indigo.server.log("Device properties for %s" % dev.name)
                        indigo.server.log(str(response))
                    self.sleep(2)
                    today = datetime.datetime.now()
                    queryY = today.strftime("%Y")
                    queryM = today.strftime("%m")
                    queryD = today.strftime("%d")
                    queryH = today.strftime("%H")
                    cMonth = queryY+queryM
                    cDay = cMonth+queryD
                    cHour = cDay+queryH
                    responseM = self.openapi.get("/v1.0/devices/{devId}/statistics/months?code=add_ele&end_month={eMonth}&start_month={sMonth}&stat_type=sum".format(devId = dev.address, eMonth = cMonth, sMonth = cMonth))
                    self.sleep(2)
                    responseD = self.openapi.get("/v1.0/devices/{devId}/statistics/days?code=add_ele&end_day={eDay}&start_day={sDay}&stat_type=sum".format(devId = dev.address, eDay = cDay, sDay = cDay))
                    self.sleep(2)
                    responseH = self.openapi.get("/v1.0/devices/{devId}/statistics/hours?code=add_ele&end_hour={eHour}&start_hour={sHour}&stat_type=sum".format(devId = dev.address, eHour = cDay+"23", sHour = cDay+"00"))
                    if self.logComms:
                        indigo.server.log("****************************************************")
                        indigo.server.log("Device statistics for %s" % dev.name)
                        indigo.server.log(str(responseM))
                        indigo.server.log(str(responseD))
                        indigo.server.log(str(responseH))

                    try:
                        energyM = responseM["result"]["months"][cMonth]
                        energyD = responseD["result"]["days"][cDay]
                        energyH = responseH["result"]["hours"][cHour]
                    except:
                        indigo.server.log("Cannot retrieve stats for %s" % dev.name, isError=True)
                        continue
                    energy = 0
                    current = 0
                    power = 0
                    voltage = 0
                    data_errors_p = 0
                    try:
                        if response["success"]:
                            for i in range(0, len(response["result"]["properties"])):
                                vName = response["result"]["properties"][i]["code"]
                                vValue = response["result"]['properties'][i]["value"]
                                if "add_ele" in vName:
                                    energy = round((float(vValue) / 1000), 2)
                                elif "cur_current" in vName:
                                    current = round((float(vValue) / 1000), 1)
                                elif "cur_power" in vName:
                                    power = float(vValue) / 10
                                elif "cur_voltage" in vName:
                                    voltage = float(vValue) / 10
                            key_value_list = [
                                {'key':'energy_this_hour', 'value':energyH, 'uiValue':str(energyH)+" kWh", 'decimalPlaces':2},
                                {'key':'energy_this_day', 'value':energyD, 'uiValue':str(energyD)+" kWh", 'decimalPlaces':1},
                                {'key':'energy_this_month', 'value':energyM, 'uiValue':str(energyM)+" kWh", 'decimalPlaces':1},
                                {'key':'cur_current', 'value':current, 'uiValue':str(current)+" A", 'decimalPlaces':1},
                                {'key':'cur_power', 'value':power, 'uiValue':str(power)+" W", 'decimalPlaces':1},
                                {'key':'cur_voltage', 'value':voltage, 'uiValue':str(voltage)+" V", 'decimalPlaces':1},
                                {'key':'state', 'value':"GOOD", 'uiValue':"GOOD"}
                            ]
                            self.powerRetries = 0
                        else:
                            key_value_list.append({'key':'state', 'value':"BAD", 'uiValue':"BAD"})
                    except:
                        # Retries code added because this call fails occasionally
                        self.powerRetries += 1
                        if self.powerRetries == 3:
                            indigo.server.log("Cannot retrieve properties for %s" % dev.name, isError=True)
                            self.powerRetries = 0
                        continue

                elif dev.deviceTypeId == "tuyaPower150":
                    try:     
                        responseP = self.openapi.get("/v2.0/cloud/thing/{}/shadow/properties".format(dev.address))
                    except:
                        indigo.server.log("Error accessing Tuya servers for Smart Power Monitor")
                        continue
                    if self.logComms:
                        indigo.server.log("----------------------------------------------------")
                        indigo.server.log("Device properties for %s" % dev.name)
                        indigo.server.log(str(responseP))
                    today = datetime.date.today()
                    strYear = today.strftime("%y")
                    decYear = int(strYear)
                    hexYear = f'{decYear:02X}'
                    strMonth = today.strftime("%m")
                    decMonth = int(strMonth)
                    hexMonth = f'{decMonth:02X}'
                    strDay = today.strftime("%d")
                    decDay = int(strDay)
                    hexDay = f'{decDay:02X}'
                    dRange = self.b64e(hexMonth+hexDay+hexMonth+hexDay)
                    # Force update of daily energy total
                    data = {"commands":[{"code":"energy_daily","value":dRange}]}
                    responseCd = self.openapi.post("/v1.0/devices/{}/commands".format(dev.address), data)    # This is a command; we don't need the response
                    self.sleep(2)
                    mRange = self.b64e(hexYear+hexMonth+hexYear+hexMonth)
                    # Force update of monthly energy total
                    data = {"commands":[{"code":"energy_month","value":mRange}]}
                    responseCm = self.openapi.post("/v1.0/devices/{}/commands".format(dev.address), data)    # This is a command; we don't need the response
                    mEnergy = 0
                    dEnergy = 0
                    current = 0
                    power = 0
                    voltage = 0
                    curDir = "FORWARD"
                    key_value_list = []
                    try:
                        data_errors = 0
                        if response["success"]:
                            for i in range(0, len(response["result"])):
                                vName = response["result"][i]["code"]
                                vValue = response["result"][i]["value"]
                                if "phase_a" in vName:
                                    voltageRaw = int(self.b64d(vValue)[:4], 16)
                                    currentRaw = int(self.b64d(vValue)[4:10], 16)
                                    activePowerRaw = int(self.b64d(vValue)[10:], 16)
                                    current = round(float(currentRaw) / 1000, 2)
                                    voltage = round(float(voltageRaw) / 10, 1)
                                    activePower = round(float(activePowerRaw) / 1000, 3)
                                    key_value_list.append({'key':'electricCurrent', 'value':current, 'uiValue':str(current)+" A", 'decimalPlaces':2})
                                    key_value_list.append({'key':'cur_voltage', 'value':voltage, 'uiValue':str(voltage)+" V", 'decimalPlaces':1})
                                    key_value_list.append({'key':'activePower', 'value':activePower, 'uiValue':str(activePower)+" kW", 'decimalPlaces':3})
                                elif "energy_month" in vName:
                                    mEnergyRaw = int(self.b64d(vValue)[8:], 16)
                                    mEnergy = round(float(mEnergyRaw) / 100, 2)
                                    key_value_list.append({'key':'month_energy', 'value':mEnergy, 'uiValue':str(mEnergy)+" kWh", 'decimalPlaces':2})
                                elif "energy_daily" in vName:
                                    dEnergyRaw = int(self.b64d(vValue)[8:], 16)
                                    dEnergy = round(float(dEnergyRaw) / 100, 2) 
                                    key_value_list.append({'key':'daily_energy', 'value':dEnergy, 'uiValue':str(dEnergy)+" kWh", 'decimalPlaces':2})
                            if responseP["success"]:                         
                                for i in range(0, len(responseP["result"]["properties"])):
                                    vName = responseP["result"]["properties"][i]["code"]
                                    try:
                                        vValue = responseP["result"]['properties'][i]["value"]
                                    except:
                                        vValue = "0"
                                    if "current_direction" in vName:
                                        curDir = vValue
                                        key_value_list.append({'key':'current_direction', 'value':curDir, 'uiValue':str(curDir)})                          
                                for i in range(0, len(responseP["result"]["properties"])):
                                    vName = responseP["result"]["properties"][i]["code"]
                                    try:
                                        vValue = responseP["result"]['properties'][i]["value"]
                                    except:
                                        vValue = "0"
                                    if "power" in vName:
                                        power = float(vValue)
                                        if str(curDir) == "REVERSE":
                                            power = power * -1
                                        key_value_list.append({'key':'cur_power', 'value':power, 'uiValue':str(power)+" W", 'decimalPlaces':1})
                                    if "forward_energy_total" in vName:
                                        feTotal = round((float(vValue) / 100), 2)
                                        key_value_list.append({'key':'forward_energy_total', 'value':feTotal, 'uiValue':str(feTotal)+" kWh", 'decimalPlaces':2})
                                    elif "reverse_energy_total" in vName:
                                        reTotal = round((float(vValue) / 100), 2)
                                        key_value_list.append({'key':'reverse_energy_total', 'value':reTotal, 'uiValue':str(reTotal)+" kWh", 'decimalPlaces':2})
                                self.power150Retries = 0
                            else:
                                data_errors += 1
                        else:
                            data_errors +=1
                        if data_errors == 0:
                            key_value_list.append({'key':'state', 'value':"GOOD", 'uiValue':"GOOD"})
                        else:
                            key_value_list.append({'key':'state', 'value':"BAD", 'uiValue':"BAD"})
                        
                    except:
                        # Retries code added because this call fails occasionally
                        self.power150Retries += 1
                        if self.power150Retries == 3:
                            indigo.server.log("Cannot retrieve properties for %s" % dev.name, isError=True)
                            self.power150Retries = 0
                        continue                    
                                        
                elif dev.deviceTypeId == "tuyaKeypad":
                    try:     
                        response = self.openapi.get("/v2.0/cloud/thing/{}/shadow/properties".format(dev.address))
                    except:
                        indigo.server.log("Error accessing Tuya servers for Smart Keypad/Reader")
                        continue
                    if self.logComms:
                        indigo.server.log("----------------------------------------------------")
                        indigo.server.log("Device properties for %s" % dev.name)
                        indigo.server.log(str(response))
                    
                elif dev.deviceTypeId == "tuyaFan":
                    try:     
                        responseF = self.openapi.get("/v2.0/cloud/thing/{}/shadow/properties".format(dev.address))
                    except:
                        indigo.server.log("Error accessing Tuya servers for Smart Fan", isError=True)
                        continue
                    if self.logComms:
                        indigo.server.log("----------------------------------------------------")
                        indigo.server.log("Device properties for %s" % dev.name)
                        indigo.server.log(str(responseF))
                    if responseF["success"]:
                        fanSpeed = 0
                        fanDir = ""
                        fanCount = 0                      
                        for i in range(0, len(responseF["result"]["properties"])):
                            vName = responseF["result"]["properties"][i]["code"]
                            try:
                                vValue = responseF["result"]['properties'][i]["value"]
                            except:
                                vValue = "0"
                            if "fan_switch" in vName:
                                onState = vValue
                                if onState:
                                    dev.updateStateOnServer(key="onOffState", value=True, clearErrorState=True)
                                    dev.updateStateImageOnServer(indigo.kStateImageSel.PowerOn)
                                else:
                                    dev.updateStateOnServer(key="onOffState", value=False, clearErrorState=True)
                                    dev.updateStateImageOnServer(indigo.kStateImageSel.PowerOff)
                            elif "fan_speed" in vName:
                                fanSpeed = int(vValue)
                            elif "fan_direction" in vName:
                                fanDir = vValue
                            elif "countdown_left_fan" in vName:
                                fanCount = int(vValue)     
                        key_value_list = [
                            {'key':'fan_speed', 'value':fanSpeed, 'uiValue':str(fanSpeed)},
                            {'key':'fan_direction', 'value':fanDir, 'uiValue':str(fanDir)},
                            {'key':'countdown_left_fan', 'value':fanCount, 'uiValue':str(fanCount)}
                        ]
                    else:
                        indigo.server.log("Cannot retrieve properties for Smart Fan", isError=True)
                        continue                

                elif dev.deviceTypeId == "tuyaEvCharger":
                    try:     
                        responseF = self.openapi.get("/v2.0/cloud/thing/{}/shadow/properties".format(dev.address))
                    except:
                        indigo.server.log("Error accessing Tuya servers for Smart EV Charger")
                        continue
                    if self.logComms:
                        indigo.server.log("----------------------------------------------------")
                        indigo.server.log("Device properties for %s" % dev.name)
                        indigo.server.log(str(responseF))
                    energy = 0
                    battery = 0
                    key_value_list = []
                    try:
                        if responseF["success"]:                            
                            for i in range(0, len(responseF["result"]["properties"])):
                                vName = responseF["result"]["properties"][i]["code"]
                                try:
                                    vValue = responseF["result"]['properties'][i]["value"]
                                except:
                                    vValue = "0"
                                if "forward_energy_total" in vName:
                                    fet = round(float(int(vValue) / 100), 3)
                                    key_value_list.append({'key':'forward_energy_total', 'value':fet, 'uiValue':str(fet)+" kWh"})
                                elif "battery_percentage" in vName:
                                    bp = int(vValue)
                                    key_value_list.append({'key':'battery_percentage', 'value':bp, 'uiValue':str(bp)+"%"})
                                elif "work_state" in vName:
                                    ws = vValue
                                    key_value_list.append({'key':'work_state', 'value':ws, 'uiValue':ws})
                                elif "work_mode" in vName:
                                    wMode = vValue
                                    key_value_list.append({'key':'work_mode', 'value':wMode, 'uiValue':wMode})
                                elif "charge_cur_set" in vName:
                                    ccs = int(vValue)
                                    key_value_list.append({'key':'charge_cur_set', 'value':ccs, 'uiValue':str(ccs)+" A"})
                                elif "phase_a" in vName:
                                    voltageRaw = int(self.b64d(vValue)[:4], 16)
                                    currentRaw = int(self.b64d(vValue)[4:10], 16)
                                    activePowerRaw = int(self.b64d(vValue)[10:], 16)
                                    current = round(float(currentRaw) / 1000, 3)
                                    voltage = round(float(voltageRaw) / 10, 1)
                                    activePower = round(float(activePowerRaw), 3)
                                    key_value_list.append({'key':'current', 'value':current, 'uiValue':str(current)+" A", 'decimalPlaces':3})
                                    key_value_list.append({'key':'voltage', 'value':voltage, 'uiValue':str(voltage)+" V", 'decimalPlaces':1})
                                    key_value_list.append({'key':'active_power', 'value':activePower, 'uiValue':str(activePower)+" W"})
                                elif "connection_state" in vName:
                                    cs = vValue
                                    key_value_list.append({'key':'connection_state', 'value':cs, 'uiValue':cs}) 
                                elif "meter_id" in vName:
                                    mid = vValue
                                    key_value_list.append({'key':'meter_id', 'value':mid, 'uiValue':mid})
                                elif "mode" in vName:
                                    mode = vValue
                                    key_value_list.append({'key':'mode', 'value':mode, 'uiValue':mode})
                                elif "occp_ip" in vName:
                                    occpIp = vValue
                                    key_value_list.append({'key':'occp_ip', 'value':occpIp, 'uiValue':occpIp})                         
                                elif "occp_port" in vName:
                                    occpPort = vValue
                                    {'key':'occp_port', 'value':occpPort, 'uiValue':occpPort}                     
                                elif "system_version" in vName:
                                    sv = vValue
                                    key_value_list.append({'key':'system_version', 'value':sv, 'uiValue':sv})
        #                         else:
        #                             indigo.server.log(str(vName)+": "+str(vValue))
                            self.evRetries = 0
                    except:
                        # Retries code added because this call fails occasionally
                        self.evRetries += 1
                        if self.evRetries == 3:
                            indigo.server.log("Cannot retrieve properties for %s" % dev.name, isError=True)
                            self.evRetries = 0
                        continue

                elif dev.deviceTypeId == "tuyaIntercom":
                    try:     
                        responseF = self.openapi.get("/v2.0/cloud/thing/{}/shadow/properties".format(dev.address))
                    except:
                        indigo.server.log("Error accessing Tuya servers for Intercom")
                        continue
                    if self.logComms:
                        indigo.server.log("----------------------------------------------------")
                        indigo.server.log("Device properties for %s" % dev.name)
                        indigo.server.log(str(responseF))                   
                    for i in range(0, len(responseF["result"]["properties"])):
                        vName = responseF["result"]["properties"][i]["code"]
                        try:
                            vValue = responseF["result"]['properties'][i]["value"]
                        except:
                            vValue = "0"
                        if "doorbell_event" in vName:
                            devent = vValue 
                    key_value_list = [
                        {'key':'doorbell_event', 'value':devent, 'uiValue':str(devent)}
                    ]

                elif dev.deviceTypeId == "tuyaWaterController":
                    try:     
                        responseF = self.openapi.get("/v2.0/cloud/thing/{}/shadow/properties".format(dev.address))
                    except:
                        indigo.server.log("Error accessing Tuya servers for Water Controller")
                        continue
                    if self.logComms:
                        indigo.server.log("----------------------------------------------------")
                        indigo.server.log("Device properties for %s" % dev.name)
                        indigo.server.log(str(responseF))                  
                    try:
                        for i in range(0, len(responseF["result"]["properties"])):
                            vName = responseF["result"]["properties"][i]["code"]
                            try:
                                vValue = responseF["result"]['properties'][i]["value"]
                            except:
                                vValue = "0"
                            if "ph_current" in vName:
                                ph_cur = round(float(int(vValue) / 100), 2)
                            elif "ph_warn_max" in vName:
                                ph_max = round(float(int(vValue) / 100), 1)
                            elif "ph_warn_min" in vName:
                                ph_min = round(float(int(vValue) / 100), 1)
                            elif "ec_current" in vName:
                                ec_cur = int(vValue)
                            elif "ec_warn_max" in vName:
                                ec_max = int(vValue)
                            elif "ec_warn_min" in vName:
                                ec_min = int(vValue)
                            elif "temp_current" in vName:
                                temp_cur = round(float(int(vValue) / 10), 1)
                            elif "temp_warn_max" in vName:
                                temp_max = round(float(int(vValue) / 10), 1)
                            elif "temp_warn_min" in vName:
                                temp_min = round(float(int(vValue) / 10), 1)
                            
    #                        else:
    #                            indigo.server.log(str(vName)+": "+str(vValue))
                        key_value_list = [
                            {'key':'ph_current', 'value':ph_cur, 'uiValue':str(ph_cur) },
                            {'key':'ph_warn_max', 'value':ph_max, 'uiValue':str(ph_max) },
                            {'key':'ph_warn_min', 'value':ph_min, 'uiValue':ph_min},
                            {'key':'ec_current', 'value':ec_cur, 'uiValue':str(ec_cur)+" us/cm" },
                            {'key':'ec_warn_max', 'value':ec_max, 'uiValue':str(ec_max)+" us/cm" },
                            {'key':'ec_warn_min', 'value':ec_min, 'uiValue':str(ec_min)+" us/cm" },
                            {'key':'temp_current', 'value':temp_cur, 'uiValue':str(temp_cur)+" °C" },
                            {'key':'temp_warn_max', 'value':temp_max, 'uiValue':str(temp_max)+" °C" },
                            {'key':'temp_warn_min', 'value':temp_min, 'uiValue':str(temp_min)+" °C" },
                            ]
                        self.wcRetries = 0
                    except:
                        # Retries code added because this call fails occasionally
                        self.wcRetries += 1
                        if self.wcRetries == 3:
                            indigo.server.log("Cannot retrieve properties for %s" % dev.name, isError=True)
                            self.wcRetries = 0
                        continue

                else:
                    # Updates for this device type not supported
                    continue
                        
                dev.updateStatesOnServer(key_value_list, clearErrorState=False)
                
            elif response["msg"] == "token invalid":
                indigo.server.log("Refreshing API token for readings")
                self.openapi.get_token()
                
            else:
                indigo.server.log("Cannot retrieve status for %s" % dev.name, isError=True)
                indigo.server.log(str(response))

            # Get info
            response = self.openapi.get("/v1.1/iot-03/devices/{}".format(dev.address))
            if response["success"] == True:
                if self.logComms:
                    indigo.server.log("----------------------------------------------------")
                    indigo.server.log("Device info for %s" % dev.name)
                    indigo.server.log(str(response))    
                local_key = response["result"]["local_key"]
                category = response["result"]["category_name"]
                model = response["result"]["model"]
                name = response["result"]["name"]
                ip = response["result"]["ip"]
                product = response["result"]["product_name"]
                if response["result"]["online"] == True:
                    status = "Online"
                else:
                    status = "Offline"
        
                key_value_list = [
                    {'key':'local_key', 'value':local_key},
                    {'key':'category', 'value':category},
                    {'key':'model', 'value':model},
                    {'key':'name', 'value':name},
                    {'key':'product', 'value':product},
                    {'key':'status', 'value':status},
                    {'key':'ip_address', 'value':ip},
                ]
                dev.updateStatesOnServer(key_value_list, clearErrorState=False)
            elif response["msg"] == "token invalid":
                indigo.server.log("Refreshing API token for info")
                self.openapi.get_token()  
            else:
                indigo.server.log("Error getting info for %s" % dev.name, isError=True)
                indigo.server.log(str(response))


    def b64e(self, s):
        return base64.b64encode(bytes.fromhex(s)).decode()
    
    
    def b64d(self, s):
        return base64.b64decode(s.encode()).hex()
        


    #######################################
    # Action functions
    #######################################
    
    def setWhiteMode(self, pluginAction, actionId):
        devToProcess = int(pluginAction.props["bulb"])
        device = None
        for dev in indigo.devices.iter("self"):
            if dev.id == devToProcess:
                device = indigo.devices[dev.name]
                if device != None:
                    if dev.deviceTypeId == "tuyaBulb":
                        data = {"commands":[{"code":"work_mode","value":"white"}]}
                        try:
                            response = self.openapi.post("/v1.0/iot-03/devices/{}/commands".format(dev.address), data)
                            if response["success"] == True:
                                dev.updateStateOnServer("mode", "white")
                                return True
                            else:
                                indigo.server.log("%s mode -> white command failed" % dev.name, isError=True)
                        except:
                            indigo.server.log("%s mode -> white command failed" % dev.name, isError=True)          
                    else:
                        indigo.server.log("Colour commands can only be sent to colour bulb devices", isError=True)
            

    def setColourMode(self, pluginAction, actionId):
        devToProcess = int(pluginAction.props["bulb"])
        device = None
        for dev in indigo.devices.iter("self"):
            if dev.id == devToProcess:
                device = indigo.devices[dev.name]
                if device != None:
                    if dev.deviceTypeId == "tuyaBulb":
                        data = {"commands":[{"code":"work_mode","value":"colour"}]}
                        try:
                            response = self.openapi.post("/v1.0/iot-03/devices/{}/commands".format(dev.address), data)
                            if response["success"] == True:
                                dev.updateStateOnServer("mode", "colour")
                                return True
                            else:
                                indigo.server.log("%s mode -> colour command failed" % dev.name, isError=True)
                        except:
                            indigo.server.log("%s mode -> colour command failed" % dev.name, isError=True)          
                    else:
                        indigo.server.log("Colour commands can only be sent to colour bulb devices", isError=True)  


    def sendIrCommand(self, pluginAction):
        ir_transmitter_device = indigo.devices[int(pluginAction.props["ir_transmitter"])]
        ir_transmitter = ir_transmitter_device.address
        ir_remote_device = indigo.devices[int(pluginAction.props["ir_remote"])]
        ir_remote = ir_remote_device.address
        ir_command = pluginAction.props["ir_command"]
        indigo.server.log("Sending IR command %s" % ir_command)
        data = {"categoryId": "infrared_tv", "key": ir_command}
        if ir_command == "mute":
        	raw = "raw/"
        else:
        	raw = ""
        try:
            response = self.openapi.post("/v2.0/infrareds/{}/remotes/{}/{}command".format(ir_transmitter, ir_remote, raw), data)
            if not response['success']:
                indigo.server.log("Command error", isError=True)
                indigo.server.log(str(response))
            else:
                return True
        except:
            indigo.server.log("Command fail", isError=True)
            indigo.server.log(str(response))


    def startVac(self, pluginAction):
        devToProcess = int(pluginAction.props["vac"])
        device = None
        for dev in indigo.devices.iter("self"):
            if dev.id == devToProcess:
                device = indigo.devices[dev.name]
                if device != None:
                    if dev.deviceTypeId == "tuyaVac":
                        data = {"commands":[{"code":"mode","value":"smart"}]}
                        try:
                            response = self.openapi.post("/v1.0/iot-03/devices/{}/commands".format(dev.address), data)
                            if response["success"] == True:
                                dev.updateStateOnServer("mode", "smart")
                            else:
                                indigo.server.log("%s mode -> smart command error" % dev.name, isError=True)
                        except:
                            indigo.server.log("%s mode -> smart command failed" % dev.name, isError=True)
                        self.sleep(2)        
                        data = {"commands":[{"code":"power_go","value":True}]}
                        try:
                            response = self.openapi.post("/v1.0/iot-03/devices/{}/commands".format(dev.address), data)
                            if response["success"] == True:
                                dev.updateStateOnServer("power_go", True)
                                return True
                            else:
                                indigo.server.log("%s power_go -> true command error" % dev.name, isError=True)
                        except:
                            indigo.server.log("%s power_go -> true command failed" % dev.name, isError=True)
                    else:
                        indigo.server.log("Start commands can only be sent to smart vacuum devices", isError=True)      


    def stopVac(self, pluginAction):
        devToProcess = int(pluginAction.props["vac"])
        device = None
        for dev in indigo.devices.iter("self"):
            if dev.id == devToProcess:
                device = indigo.devices[dev.name]
                if device != None:
                    if dev.deviceTypeId == "tuyaVac":
                        data = {"commands":[{"code":"mode","value":"smart"}]}
                        try:
                            response = self.openapi.post("/v1.0/iot-03/devices/{}/commands".format(dev.address), data)
                            if response["success"] == True:
                                dev.updateStateOnServer("mode", "smart")
                            else:
                                indigo.server.log("%s mode -> smart command error" % dev.name, isError=True)
                        except:
                            indigo.server.log("%s mode -> smart command failed" % dev.name, isError=True)
                        self.sleep(2)        
                        data = {"commands":[{"code":"power_go","value":False}]}
                        try:
                            response = self.openapi.post("/v1.0/iot-03/devices/{}/commands".format(dev.address), data)
                            if response["success"] == True:
                                dev.updateStateOnServer("power_go", False)
                                return True
                            else:
                                indigo.server.log("%s power_go -> false command error" % dev.name, isError=True)
                        except:
                            indigo.server.log("%s power_go -> false command failed" % dev.name, isError=True)
                    else:
                        indigo.server.log("Stop commands can only be sent to smart vacuum devices", isError=True)      
        
 
    def chargeVac(self, pluginAction):
        devToProcess = int(pluginAction.props["vac"])
        device = None
        for dev in indigo.devices.iter("self"):
            if dev.id == devToProcess:
                device = indigo.devices[dev.name]
                if device != None:
                    if dev.deviceTypeId == "tuyaVac":
                        data = {"commands":[{"code":"mode","value":"chargego"}]}
                        try:
                            response = self.openapi.post("/v1.0/iot-03/devices/{}/commands".format(dev.address), data)
                            if response["success"] == True:
                                dev.updateStateOnServer("mode", "chargego")
                                return True
                            else:
                                indigo.server.log("%s mode -> charge command error" % dev.name, isError=True)
                        except:
                            indigo.server.log("%s mode -> charge command failed" % dev.name, isError=True)
                    else:
                        indigo.server.log("Charge commands can only be sent to smart vacuum devices", isError=True)
                        

    def fanForward(self, pluginAction):
        devToProcess = int(pluginAction.props["fan"])
        device = None
        for dev in indigo.devices.iter("self"):
            if dev.id == devToProcess:
                device = indigo.devices[dev.name]
                if device != None:
                    if dev.deviceTypeId == "tuyaFan":
                        data = {"commands":[{"code":"fan_direction","value":"forward"}]}
                        try:
                            response = self.openapi.post("/v1.0/iot-03/devices/{}/commands".format(dev.address), data)
                            if response["success"] == True:
                                return True
                            else:
                                indigo.server.log("%s forward command error" % dev.name, isError=True)
                        except:
                            indigo.server.log("%s forward command failed" % dev.name, isError=True)          
                    else:
                        indigo.server.log("Direction change commands can only be sent to smart fans", isError=True)


    def fanReverse(self, pluginAction):
        devToProcess = int(pluginAction.props["fan"])
        device = None
        for dev in indigo.devices.iter("self"):
            if dev.id == devToProcess:
                device = indigo.devices[dev.name]
                if device != None:
                    if dev.deviceTypeId == "tuyaFan":
                        data = {"commands":[{"code":"fan_direction","value":"reverse"}]}
                        try:
                            response = self.openapi.post("/v1.0/iot-03/devices/{}/commands".format(dev.address), data)
                            if response["success"] == True:
                                return True
                            else:
                                indigo.server.log("%s reverse command error" % dev.name, isError=True)
                        except:
                            indigo.server.log("%s reverse command failed" % dev.name, isError=True)          
                    else:
                        indigo.server.log("Direction change commands can only be sent to smart fans", isError=True) 
                        

    def fanSpeed(self, pluginAction):
        devToProcess = int(pluginAction.props["fan"])
        requiredSpeed = int(pluginAction.props["fspeed"])
        device = None
        for dev in indigo.devices.iter("self"):
            if dev.id == devToProcess:
                device = indigo.devices[dev.name]
                if device != None:
                    if dev.deviceTypeId == "tuyaFan":
                        data = {"commands":[{"code":"fan_speed","value":requiredSpeed}]}
                        try:
                            response = self.openapi.post("/v1.0/iot-03/devices/{}/commands".format(dev.address), data)
                            if response["success"] == True:
                                return True
                            else:
                                indigo.server.log("%s speed change command error" % dev.name, isError=True)
                        except:
                            indigo.server.log("%s speed change command failed" % dev.name, isError=True)          
                    else:
                        indigo.server.log("Speed change commands can only be sent to smart fans", isError=True)
                        

    def startCharge(self, pluginAction):
        devToProcess = int(pluginAction.props["evc"])
        device = None
        for dev in indigo.devices.iter("self"):
            if dev.id == devToProcess:
                device = indigo.devices[dev.name]
                if device != None:
                    if dev.deviceTypeId == "tuyaEvCharger":
                        data = {"commands":[{"code":"switch","value":True}]}
                        try:
                            response = self.openapi.post("/v1.0/iot-03/devices/{}/commands".format(dev.address), data)
                            if response["success"] == True:
                                return True
                            else:
                                indigo.server.log("%s start charge command error" % dev.name, isError=True)
                                indigo.server.log(str(response))
                        except:
                            indigo.server.log("%s start charge command failed" % dev.name, isError=True)          
                    else:
                        indigo.server.log("Start charge commands can only be sent to smart EV Chargers", isError=True) 
                        

    def stopCharge(self, pluginAction):
        devToProcess = int(pluginAction.props["evc"])
        device = None
        for dev in indigo.devices.iter("self"):
            if dev.id == devToProcess:
                device = indigo.devices[dev.name]
                if device != None:
                    if dev.deviceTypeId == "tuyaEvCharger":
                        data = {"commands":[{"code":"switch","value":False}]}
                        try:
                            response = self.openapi.post("/v1.0/iot-03/devices/{}/commands".format(dev.address), data)
                            if response["success"] == True:
                                return True
                            else:
                                indigo.server.log("%s stop charge command error" % dev.name, isError=True)
                                indigo.server.log(str(response))
                        except:
                            indigo.server.log("%s stop charge command failed" % dev.name, isError=True)          
                    else:
                        indigo.server.log("Stop charge commands can only be sent to smart EV Chargers", isError=True)

                       

    ########################################
    # Smart device action callback
    ########################################
    
    # Main switch action bottleneck called by Indigo Server.
    def actionControlDevice(self, action, dev):
        response = ""
        if action.deviceAction == indigo.kDeviceAction.TurnOn: 
            data = {"commands":[{"code":self.switchCommand[dev.address],"value":True}]}
            if self.logCommands:
                indigo.server.log("Sending %s to %s" % (str(data), dev.name))
            try:
                response = self.openapi.post("/v1.0/iot-03/devices/{}/commands".format(dev.address), data)
                if response["success"] == True:
                    dev.updateStateOnServer("onOffState", True)
                    dev.updateStateImageOnServer(indigo.kStateImageSel.PowerOn)
                else:
                    #indigo.server.log("%s 'on' cloud command failed; trying local control" % dev.name, isError=True)
                    #indigo.server.log(str(response))
                    if dev.states["category"] == "Socket":
                        d = tinytuya.OutletDevice(dev.address, "0.0.0.0", dev.states["local_key"])
                        d.set_version(3.1)
                    else:
                        d = tinytuya.BulbDevice(dev.address, "0.0.0.0", dev.states["local_key"])
                        d.set_version(3.3)
                        d.turn_on()
                        dev.updateStateOnServer("onOffState", True)
                        dev.updateStateImageOnServer(indigo.kStateImageSel.PowerOn)
                        d.close()
            except:
                indigo.server.log("%s 'on' command failed" % dev.name, isError=True)
                

        elif action.deviceAction == indigo.kDeviceAction.TurnOff:
            data = {"commands":[{"code":self.switchCommand[dev.address],"value":False}]}
            if self.logCommands:
                indigo.server.log("Sending %s to %s" % (str(data), dev.name))
            try:
                response = self.openapi.post("/v1.0/iot-03/devices/{}/commands".format(dev.address), data)
                if response["success"] == True:
                    dev.updateStateOnServer("onOffState", False)
                    dev.updateStateImageOnServer(indigo.kStateImageSel.PowerOff)
                else:
                    #indigo.server.log("%s 'off' cloud command failed, trying local control" % dev.name, isError=True)
                    #indigo.server.log(str(response))
                    if dev.states["category"] == "Socket":
                        d = tinytuya.OutletDevice(dev.address, "0.0.0.0", dev.states["local_key"])
                        d.set_version(3.1)
                    else:
                        d = tinytuya.BulbDevice(dev.address, "0.0.0.0", dev.states["local_key"])
                        d.set_version(3.3)
                        d.turn_off()
                        dev.updateStateOnServer("onOffState", False)
                        dev.updateStateImageOnServer(indigo.kStateImageSel.PowerOn)
                        d.close()
            except:
                indigo.server.log("%s 'off' command failed" % dev.name, isError=True)

                
        elif action.deviceAction == indigo.kDeviceAction.Toggle:
            originalState = dev.onState
            if dev.onState:
                data = {"commands":[{"code":self.switchCommand[dev.address],"value":False}]}
            else:
                data = {"commands":[{"code":self.switchCommand[dev.address],"value":True}]}
            if self.logCommands:
                indigo.server.log("Sending %s to %s" % (str(data), dev.name))
            try:
                response = self.openapi.post("/v1.0/iot-03/devices/{}/commands".format(dev.address), data)
                if response["success"] == True:
                    if originalState:
                        dev.updateStateOnServer("onOffState", False)
                        dev.updateStateImageOnServer(indigo.kStateImageSel.PowerOff)
                    else:
                        dev.updateStateOnServer("onOffState", True)
                        dev.updateStateImageOnServer(indigo.kStateImageSel.PowerOn)
                else:
                    indigo.server.log("%s 'toggle' command failed" % dev.name, isError=True)
                    indigo.server.log(str(response))
            except:
                indigo.server.log("%s 'toggle' command error" % dev.name, isError=True)
                indigo.server.log(str(response))

        elif action.deviceAction == indigo.kDeviceAction.SetBrightness:
            if dev.deviceTypeId == "tuyaBulb":
                new_brightness = action.actionValue
                if new_brightness < 1:
                    new_brightness = 1
                scaled_brightness = new_brightness * 10
                data = {"commands":[{"code":"bright_value_v2","value":scaled_brightness}]}
                if self.logCommands:
                    indigo.server.log("Sending %s to %s" % (str(data), dev.name))
                try:
                    response = self.openapi.post("/v1.0/iot-03/devices/{}/commands".format(dev.address), data)
                    if response["success"] == True:
                        dev.updateStateOnServer(key="brightnessLevel", value=new_brightness, uiValue=str(new_brightness)+"%")
                    else:
                        indigo.server.log(f"Command \"{dev.name}\" set brightness to {new_brightness} failed", isError=True)
                        indigo.server.log(str(response))
                except:
                    indigo.server.log(f"Command \"{dev.name}\" set brightness to {new_brightness} error", isError=True)
            else:
                indigo.server.log("Brightness commands can only be sent to bulb devices", isError=True)

        elif action.deviceAction == indigo.kDeviceAction.BrightenBy:
            if dev.deviceTypeId == "tuyaBulb":
                new_brightness = min(dev.brightness + action.actionValue, 100)
                if new_brightness < 1:
                    new_brightness = 1
                scaled_brightness = new_brightness * 10
                data = {"commands":[{"code":"bright_value_v2","value":scaled_brightness}]}
                if self.logCommands:
                    indigo.server.log("Sending %s to %s" % (str(data), dev.name))
                try:
                    response = self.openapi.post("/v1.0/iot-03/devices/{}/commands".format(dev.address), data)
                    if response["success"] == True:
                        dev.updateStateOnServer(key="brightnessLevel", value=new_brightness, uiValue=str(new_brightness)+"%")
                    else:
                        indigo.server.log(f"Command \"{dev.name}\" brighten to {new_brightness} failed", isError=True)
                        indigo.server.log(str(response))
                except:
                    indigo.server.log(f"Command \"{dev.name}\" brighten to {new_brightness} error", isError=True)
            else:
                indigo.server.log("Brightness commands can only be sent to bulb devices", isError=True)

        elif action.deviceAction == indigo.kDeviceAction.DimBy:
            if dev.deviceTypeId == "tuyaBulb":
                new_brightness = max(dev.brightness - action.actionValue, 0)
                if new_brightness < 1:
                    new_brightness = 1
                scaled_brightness = new_brightness * 10
                data = {"commands":[{"code":"bright_value_v2","value":scaled_brightness}]}
                if self.logCommands:
                    indigo.server.log("Sending %s to %s" % (str(data), dev.name))
                try:
                    response = self.openapi.post("/v1.0/iot-03/devices/{}/commands".format(dev.address), data)
                    if response["success"] == True:
                        dev.updateStateOnServer(key="brightnessLevel", value=new_brightness, uiValue=str(new_brightness)+"%")
                    else:
                        indigo.server.log(f"Command \"{dev.name}\" dim to {new_brightness} failed", isError=True)
                        indigo.server.log(str(response))
                except:
                    indigo.server.log(f"Command \"{dev.name}\" dim to {new_brightness} error", isError=True)
            else:
                indigo.server.log("Brightness commands can only be sent to bulb devices", isError=True)
                

        elif action.deviceAction == indigo.kDimmerRelayAction.SetColorLevels:
            if dev.deviceTypeId == "tuyaBulb":
                isError = False
                if dev.pluginProps.get('SupportsRGB', False):
                    if "blueLevel" in str(action.actionValue):
                        red = action.actionValue.get('redLevel', 0)
                        if red == "":
                            red = 0
                            valuesDict['redLevel'] = red
                        green = action.actionValue.get('greenLevel', 0)
                        if green == "":
                            green = 0
                            valuesDict['greenLevel'] = green
                        blue = action.actionValue.get('blueLevel', 0)
                        if blue == "":
                            blue = 0
                            valuesDict['blueLevel'] = blue

                        # Validate red value.
                        try:
                            red = int(red)
                            if (red < 0) or (red > 100):
                                isError = True
                                errorsDict['red'] = "Red values must be a whole number between 0 and 100."
                                errorsDict['showAlertText'] += errorsDict['red'] + "\n\n"
                        except ValueError:
                            isError = True
                            errorsDict['red'] = "Red values must be a whole number between 0 and 100."
                            errorsDict['showAlertText'] += errorsDict['red'] + "\n\n"
                        except Exception as e:
                            isError = True
                            errorsDict['red'] = "Invalid Red value: {}  @line#:{})".format(e,sys.exc_info()[2].tb_lineno)
                            errorsDict['showAlertText'] += errorsDict['red'] + "\n\n"

                        # Validate green value.
                        try:
                            green = int(green)
                            if (green < 0) or (green > 100):
                                isError = True
                                errorsDict['green'] = "Green values must be a whole number between 0 and 100."
                                errorsDict['showAlertText'] += errorsDict['green'] + "\n\n"
                        except ValueError:
                            isError = True
                            errorsDict['green'] = "Green values must be a whole number between 0 and 100."
                            errorsDict['showAlertText'] += errorsDict['green'] + "\n\n"
                        except Exception as e:
                            isError = True
                            errorsDict['green'] = "Invalid Green value: {}  @line#:{})".format(e,sys.exc_info()[2].tb_lineno)
                            errorsDict['showAlertText'] += errorsDict['green'] + "\n\n"

                        # Validate blue value.
                        try:
                            blue = int(blue)
                            if (blue < 0) or (blue > 100):
                                isError = True
                                errorsDict['blue'] = "Blue values must be a whole number between 0 and 100."
                                errorsDict['showAlertText'] += errorsDict['blue'] + "\n\n"
                        except ValueError:
                            isError = True
                            errorsDict['blue'] = "Blue values must be a whole number between 0 and 100."
                            errorsDict['showAlertText'] += errorsDict['blue'] + "\n\n"
                        except Exception as e:
                            isError = True
                            errorsDict['blue'] = "Invalid Blue value: {}  @line#:{})".format(e,sys.exc_info()[2].tb_lineno)
                            errorsDict['showAlertText'] += errorsDict['blue'] + "\n\n"

                        if not isError:
                            hVal, sVal, vVal = colorsys.rgb_to_hsv(red, green, blue)
                            #indigo.server.log("h: %d, s: %d, v: %d" % (hVal*360, sVal*1000, vVal*10))
                            data = {"commands":[{"code":"colour_data_v2", "value":{"h":int(hVal*360), "s":int(sVal*1000), "v":int(vVal*10)}}]}
                            if self.logCommands:
                                indigo.server.log("Sending %s to %s" % (str(data), dev.name))
                            try:
                                response = self.openapi.post("/v1.0/iot-03/devices/{}/commands".format(dev.address), data)
                                if response["success"] == True:
                                    indigo.server.log("Success!")
                                    #dev.updateStateOnServer(key="brightnessLevel", value=new_brightness, uiValue=str(new_brightness)+"%")
                                else:
                                    indigo.server.log(f"Command \"{dev.name}\" failed", isError=True)
                                    indigo.server.log(str(response))
                            except:
                                indigo.server.log(f"Command \"{dev.name}\" error", isError=True)
                                
                    elif "whiteLevel" in str(action.actionValue):
                        if dev.deviceTypeId == "tuyaBulb":
                            new_brightness = action.actionValue.get('whiteLevel', 0)
                            if new_brightness < 1:
                                new_brightness = 1
                            scaled_brightness = int(new_brightness * 10)
                            data = {"commands":[{"code":"bright_value_v2","value":scaled_brightness}]}
                            if self.logCommands:
                                indigo.server.log("Sending %s to %s" % (str(data), dev.name))
                            try:
                                response = self.openapi.post("/v1.0/iot-03/devices/{}/commands".format(dev.address), data)
                                if response["success"] == True:
                                    dev.updateStateOnServer(key="whiteLevel", value=new_brightness, uiValue=str(new_brightness)+"%")
                                else:
                                    indigo.server.log(f"Command \"{dev.name}\" set white level to {new_brightness} failed", isError=True)
                                    indigo.server.log(str(response))
                            except:
                                indigo.server.log(f"Command \"{dev.name}\" set white level to {new_brightness} error", isError=True)

                    
                    elif "whiteTemperature" in str(action.actionValue):
                        if dev.deviceTypeId == "tuyaBulb":
                            isError = False
                            whiteTemp = action.actionValue.get('whiteTemperature', 0)
                            try:
                                whiteTemp = int(whiteTemp)
                                if (whiteTemp < 2200) or (whiteTemp > 6500):
                                    isError = True
                                    errorsDict['whiteTemperature'] = "White Temperature values must be a whole number between 2200 and 6500."
                                    errorsDict['showAlertText'] += errorsDict['whiteTemperature'] + "\n\n"
                            except ValueError:
                                isError = True
                                errorsDict['whiteTemperature'] = "White Temperature values must be a whole number between 2200 and 6500."
                                errorsDict['showAlertText'] += errorsDict['whiteTemperature'] + "\n\n"
                            except Exception as e:
                                isError = True
                                errorsDict['whiteTemperature'] = "Invalid White Temperature value: {}  @line#:{})".format(e,sys.exc_info()[2].tb_lineno)
                                errorsDict['showAlertText'] += errorsDict['whiteTemperature'] + "\n\n"
                            whiteTempScaled = int((whiteTemp - 2200) / 4.3)
                            if not isError:
                                data = {"commands":[{"code":"temp_value_v2","value":whiteTempScaled}]}
                                if self.logCommands:
                                    indigo.server.log("Sending %s to %s" % (str(data), dev.name))
                                try:
                                    response = self.openapi.post("/v1.0/iot-03/devices/{}/commands".format(dev.address), data)
                                    if response["success"] == True:
                                        dev.updateStateOnServer(key="whiteTemperature", value=whiteTemp, uiValue=str(whiteTemp))
                                    else:
                                        indigo.server.log(f"Command \"{dev.name}\" set temperature to {whiteTemp} failed", isError=True)
                                        indigo.server.log(str(response))
                                except:
                                    indigo.server.log(f"Command \"{dev.name}\" set temperature to {whiteTemp} error", isError=True)
                        
            else:
                indigo.server.log("Colour commands can only be sent to colour bulb devices", isError=True)
                

